"""
offers julia-side memory management for C++ jluna
"""
module memory_handler

    const _current_id = Ref(UInt64(0));
    const _refs = Ref(Dict{UInt64, Base.RefValue{Any}}())
    const _ref_counter = Ref(Dict{UInt64, UInt64}())

    const _ref_id_marker = '#'
    const _refs_expression = Meta.parse("jluna.memory_handler._refs[]")

    # proxy id that is actually an expression, the ID of topmodule Main is
    ProxyID = Union{Expr, Symbol, Nothing}

    # make as unnamed
    make_unnamed_proxy_id(id::UInt64) = return Expr(:ref, Expr(:ref, _refs_expression, id))

    # make as named with owner and symbol name
    make_named_proxy_id(id::Symbol, owner_id::ProxyID) ::ProxyID = return Expr(:(.), owner_id, QuoteNode(id))

    # make as named with main as owner and symbol name
    make_named_proxy_id(id::Symbol, owner_id::Nothing) ::ProxyID = return id

    # make as named with owner and array index name
    make_named_proxy_id(id::Number, owner_id::ProxyID) ::ProxyID = return Expr(:ref, owner_id, convert(Int64, id))

    # assign to proxy id
    function assign(new_value::T, name::ProxyID) where T
        if new_value isa Symbol || new_value isa Expr
            return Main.eval(Expr(:(=), name, QuoteNode(new_value)))
        else
            return Main.eval(Expr(:(=), name, new_value));
        end
    end

    # eval proxy id
    evaluate(name::ProxyID) ::Any = return Main.eval(name)
    evaluate(name::Symbol) ::Any = return Main.eval(:($name))

    """
    `get_name(::ProxyID) -> String`

    parse name from proxy id
    """
    function get_name(id::ProxyID) ::String

        if length(id.args) == 0
            return "Main"
        end

        current = id
        while current.args[1] isa Expr && length(current.args) >= 2

            if current.args[2] isa UInt64
                current.args[2] = convert(Int64, current.args[2])
            end

            current = current.args[1]
        end

        out = string(id)
        reg = r"\Q((jluna.memory_handler._refs[])[\E(.*)\Q])[]\E"
        captures = match(reg, out)

        if captures != nothing
            out = replace(out, reg => "<unnamed function proxy #" * string(tryparse(Int64, captures.captures[1])) * ">")
        end

        return out;
    end

    get_name(::Nothing) ::String = return "Main"
    get_name(s::Symbol) ::String = return string(s)
    get_name(i::Integer) ::String = return "[" * string(i) * "]"

    """
    `print_refs() -> Nothing`

    pretty print _ref state, for debugging
    """
    function print_refs() ::Nothing

        println("jluna.memory_handler._refs: ");
        for e in _refs[]
            println("\t", Int64(e.first), " => ", e.second[], " (", typeof(e.second[]), ") ")
        end
    end

    """
    `create_reference(::UInt64, ::Any) -> UInt64`

    add reference to _refs
    """
    function create_reference(to_wrap::Any) ::UInt64

        global _current_id[] += 1;
        key::UInt64 = _current_id[];

        if (haskey(_refs[], key))
            _ref_counter[][key] += 1
        else
            _refs[][key] = Base.RefValue{Any}(to_wrap)
            _ref_counter[][key] = 1
        end

        return key;
    end

    create_reference(_::Nothing) ::UInt64 = return 0

    """
    `set_reference(::UInt64, ::T) -> Nothing`

    update the value of a reference in _refs without adding a new entry or changing it's key, ref pointers C++ side stay valid
    """
    function set_reference(key::UInt64, new_value::T) ::Base.RefValue{Any} where T

        _refs[][key] = Base.RefValue{Any}(new_value)
        return _refs[][key]
    end

    """
    `get_reference(::Int64) -> Any`

    access reference in _refs
    """
    function get_reference(key::UInt64) ::Any

        if (key == 0)
            return nothing
        end

       return _refs[][key]
    end

    """
    `free_reference(::UInt64) -> Nothing`

    free reference from _refs
    """
    function free_reference(key::UInt64) ::Nothing

        if (key == 0)
            return nothing;
        end

        if _refs[][key][] isa Module
            return
        end

        global _ref_counter[][key] -= 1
        count::UInt64 = _ref_counter[][key]

        if (count == 0)
            delete!(_ref_counter[], key)
            delete!(_refs[], key)
        end

        return nothing;
    end

    """
    `force_free() -> Nothing`

    immediately deallocate all C++-managed values
    """
    function force_free() ::Nothing

        for k in keys(_refs)
            delete!(_refs[], k)
            delete!(_ref_counter[], k)
        end

        return nothing;
    end
end